#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import fixtures
import mock
import os
import sys
import tempfile
import yaml

from heatclient import exc as hc_exc
from tripleo_common.image import kolla_builder

from tripleoclient import exceptions
from tripleoclient.tests import fakes
from tripleoclient.tests.v1.test_plugin import TestPluginV1

# Load the plugin init module for the plugin list and show commands
from tripleoclient.v1 import tripleo_deploy

import ansible_runner


# TODO(sbaker) Remove after a tripleo-common release contains this new function
if not hasattr(kolla_builder, 'container_images_prepare_multi'):
    setattr(kolla_builder, 'container_images_prepare_multi', mock.Mock())


class FakePluginV1Client(object):
    def __init__(self, **kwargs):
        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']


class TestDeployUndercloud(TestPluginV1):

    def setUp(self):
        super(TestDeployUndercloud, self).setUp()

        # Get the command object to test
        self.cmd = tripleo_deploy.Deploy(self.app, None)
        self.cmd.ansible_dir = '/tmp'

        tripleo_deploy.Deploy.heat_pid = mock.MagicMock(
            return_value=False)
        tripleo_deploy.Deploy.tht_render = '/twd/templates'
        tripleo_deploy.Deploy.heat_launch = mock.MagicMock(
            side_effect=(lambda *x, **y: None))

        self.tc = self.app.client_manager.tripleoclient = mock.MagicMock()
        self.orc = self.tc.local_orchestration = mock.MagicMock()
        self.orc.stacks.create = mock.MagicMock(
            return_value={'stack': {'id': 'foo'}})

    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy._is_undercloud_deploy')
    @mock.patch('tripleoclient.utils.check_hostname')
    def test_run_preflight_checks(self, mock_check_hostname, mock_uc):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8'], [])

        mock_uc.return_value = False
        self.cmd._run_preflight_checks(parsed_args)
        mock_check_hostname.called_one_with(False)

    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy._is_undercloud_deploy')
    @mock.patch('tripleoclient.utils.check_hostname')
    def test_run_preflight_checks_output_only(self, mock_check_hostname,
                                              mock_uc):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--output-only'], [])

        mock_uc.return_value = False
        self.cmd._run_preflight_checks(parsed_args)
        mock_check_hostname.assert_not_called()

    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy._is_undercloud_deploy')
    @mock.patch('tripleoclient.utils.check_hostname')
    def test_run_preflight_checks_undercloud(self, mock_check_hostname,
                                             mock_uc):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8'], [])

        mock_uc.return_value = True
        self.cmd._run_preflight_checks(parsed_args)
        mock_check_hostname.assert_not_called()

    def test_get_roles_file_path(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8'], [])

        roles_file = self.cmd._get_roles_file_path(parsed_args)
        self.assertEqual(roles_file,
                         '/usr/share/openstack-tripleo-heat-templates/'
                         'roles_data_undercloud.yaml')

    def test_get_roles_file_path_custom_file(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot',
                                         '--roles-file', 'foobar.yaml'], [])

        roles_file = self.cmd._get_roles_file_path(parsed_args)
        self.assertEqual(roles_file, 'foobar.yaml')

    def test_get_roles_file_path_custom_templates(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot'], [])

        import pprint
        pprint.pprint(parsed_args)
        roles_file = self.cmd._get_roles_file_path(parsed_args)
        self.assertEqual(roles_file,
                         '/tmp/thtroot/roles_data_undercloud.yaml')

    def test_get_networks_file_path(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8'], [])

        networks_file = self.cmd._get_networks_file_path(parsed_args)
        self.assertEqual('/dev/null', networks_file)

    def test_get_networks_file_path_custom_file(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--networks-file', 'foobar.yaml'], [])

        networks_file = self.cmd._get_networks_file_path(parsed_args)
        self.assertEqual('foobar.yaml', networks_file)

    def test_get_networks_file_path_custom_templates(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot'], [])

        networks_file = self.cmd._get_networks_file_path(parsed_args)
        self.assertEqual('/dev/null', networks_file)

    def test_get_plan_env_file_path(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8'], [])

        plan_env_file = self.cmd._get_plan_env_file_path(parsed_args)
        self.assertEqual(plan_env_file,
                         '/usr/share/openstack-tripleo-heat-templates/'
                         'plan-environment.yaml')

    def test_get_plan_env_file_path_custom_file(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot',
                                         '--plan-environment-file',
                                         'foobar.yaml'], [])

        plan_env_file = self.cmd._get_plan_env_file_path(parsed_args)
        self.assertEqual(plan_env_file, 'foobar.yaml')

    def test_get_plan_env_file_path_custom_templates(self):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot'], [])

        plan_env_file = self.cmd._get_plan_env_file_path(parsed_args)
        self.assertEqual(plan_env_file,
                         '/tmp/thtroot/plan-environment.yaml')

    @mock.patch('os.path.exists')
    @mock.patch('tripleoclient.utils.fetch_roles_file')
    def test_get_primary_role_name(self, mock_data, mock_exists):
        parsed_args = mock.Mock()
        mock_data.return_value = [
            {'name': 'Bar'}, {'name': 'Foo', 'tags': ['primary']}
        ]
        self.assertEqual(
            self.cmd._get_primary_role_name(parsed_args.roles_file,
                                            parsed_args.templates),
            'Foo')

    @mock.patch('tripleoclient.utils.fetch_roles_file', return_value=None)
    def test_get_primary_role_name_none_defined(self, mock_data):
        parsed_args = self.check_parser(self.cmd, [], [])
        self.assertEqual(
            self.cmd._get_primary_role_name(parsed_args.roles_file,
                                            parsed_args.templates),
            'Controller')

    @mock.patch('tripleoclient.utils.fetch_roles_file')
    def test_get_primary_role_name_no_primary(self, mock_data):
        parsed_args = mock.Mock()
        mock_data.return_value = [{'name': 'Bar'}, {'name': 'Foo'}]
        self.assertEqual(
            self.cmd._get_primary_role_name(parsed_args.roles_file,
                                            parsed_args.templates),
            'Bar')

    @mock.patch('os.path.exists', side_effect=[True, False])
    @mock.patch('shutil.copytree')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    def test_populate_templates_dir(self, mock_workingdirs, mock_copy,
                                    mock_exists):
        self.cmd.tht_render = '/foo'
        self.cmd._populate_templates_dir('/bar')
        mock_workingdirs.assert_called_once()
        mock_copy.assert_called_once_with('/bar', '/foo', symlinks=True)

    @mock.patch('os.path.exists', return_value=False)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    def test_populate_templates_dir_bad_source(self, mock_workingdirs,
                                               mock_exists):
        self.cmd.tht_render = '/foo'
        self.assertRaises(exceptions.NotFound,
                          self.cmd._populate_templates_dir, '/foo')

    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('getpass.getuser', return_value='stack')
    @mock.patch('os.chmod')
    @mock.patch('os.path.exists')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('subprocess.check_call', autospec=True)
    @mock.patch('tripleo_common.utils.passwords.generate_passwords')
    @mock.patch('yaml.safe_dump')
    def test_update_passwords_env_init(self, mock_dump, mock_pw, mock_cc,
                                       mock_exists, mock_chmod, mock_user):
        pw_dict = {"GeneratedPassword": 123}

        mock_pw.return_value = pw_dict
        mock_exists.return_value = False

        mock_open_context = mock.mock_open()
        with mock.patch('six.moves.builtins.open', mock_open_context):
            self.cmd._update_passwords_env(self.temp_homedir, 'stack')

        mock_open_handle = mock_open_context()
        mock_dump.assert_called_once_with({'parameter_defaults': pw_dict},
                                          mock_open_handle,
                                          default_flow_style=False)

    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('getpass.getuser', return_value='stack')
    @mock.patch('os.chmod')
    @mock.patch('os.path.exists')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('subprocess.check_call', autospec=True)
    @mock.patch('tripleo_common.utils.passwords.generate_passwords')
    @mock.patch('yaml.safe_dump')
    def test_update_passwords_env(self, mock_dump, mock_pw, mock_cc,
                                  mock_exists, mock_chmod, mock_user):
        pw_dict = {"GeneratedPassword": 123, "LegacyPass": "override me"}
        pw_conf_path = os.path.join(self.temp_homedir,
                                    'undercloud-passwords.conf')
        t_pw_conf_path = os.path.join(
            self.temp_homedir, 'tripleo-undercloud-passwords.yaml')

        mock_pw.return_value = pw_dict
        mock_exists.return_value = True
        with open(t_pw_conf_path, 'w') as t_pw:
            t_pw.write('parameter_defaults: {ExistingKey: xyz, '
                       'LegacyPass: pick-me-legacy-tht, '
                       'RpcPassword: pick-me-rpc}\n')

        with open(pw_conf_path, 'w') as t_pw:
            t_pw.write('[auth]\nundercloud_db_password = ignore-me-mysql\n'
                       'undercloud_rabbit_password = ignore-me-rabbit\n'
                       'undercloud_rpc_password = ignore-me-rpc\n'
                       'undercloud_legacy_pass = ignore-me-legacy\n')

        self.cmd._update_passwords_env(self.temp_homedir,
                                       'stack', upgrade=False,
                                       passwords={'ADefault': 456,
                                                  'ExistingKey':
                                                  'dontupdate'})
        expected_dict = {
            'parameter_defaults': {'GeneratedPassword': 123,
                                   'LegacyPass': 'pick-me-legacy-tht',
                                   'RpcPassword': 'pick-me-rpc',
                                   'ExistingKey': 'xyz',
                                   'ADefault': 456}}
        mock_dump.assert_called_once_with(expected_dict,
                                          mock.ANY,
                                          default_flow_style=False)

    # TODO(bogdando) drop once we have proper oslo.privsep
    @mock.patch('getpass.getuser', return_value='stack')
    @mock.patch('os.chmod')
    @mock.patch('os.path.exists')
    # TODO(bogdando) drop once we have proper oslo.privsep
    @mock.patch('subprocess.check_call', autospec=True)
    @mock.patch('tripleo_common.utils.passwords.generate_passwords')
    @mock.patch('yaml.safe_dump')
    def test_update_passwords_env_upgrade(self, mock_dump, mock_pw, mock_cc,
                                          mock_exists, mock_chmod, mock_user):
        pw_dict = {"GeneratedPassword": 123, "LegacyPass": "override me"}
        pw_conf_path = os.path.join(self.temp_homedir,
                                    'undercloud-passwords.conf')
        t_pw_conf_path = os.path.join(
            self.temp_homedir, 'tripleo-undercloud-passwords.yaml')

        mock_pw.return_value = pw_dict

        def mock_file_exists(file_name):
            return not file_name.startswith('/etc/keystone')
        mock_exists.side_effect = mock_file_exists
        with open(t_pw_conf_path, 'w') as t_pw:
            t_pw.write('parameter_defaults: {ExistingKey: xyz, '
                       'LegacyPass: override-me-legacy, '
                       'RpcPassword: override-me-rpc}\n')

        with open(pw_conf_path, 'w') as t_pw:
            t_pw.write('[auth]\nundercloud_db_password = pick-me-mysql\n'
                       'undercloud_rabbit_password = pick-me-rabbit\n'
                       'undercloud_rpc_password = pick-me-rpc\n'
                       'undercloud_legacy_pass = pick-me-legacy-instack\n')

        self.cmd._update_passwords_env(self.temp_homedir,
                                       'stack', upgrade=True,
                                       passwords={'ADefault': 456,
                                                  'ExistingKey':
                                                  'dontupdate'})
        expected_dict = {
            'parameter_defaults': {'GeneratedPassword': 123,
                                   'ExistingKey': 'xyz',
                                   'MysqlRootPassword': 'pick-me-mysql',
                                   'RpcPassword': 'pick-me-rpc',
                                   'RabbitPassword': 'pick-me-rabbit',
                                   'LegacyPass': 'pick-me-legacy-instack',
                                   'ADefault': 456}}
        mock_dump.assert_called_once_with(expected_dict,
                                          mock.ANY,
                                          default_flow_style=False)

    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.environment_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('heatclient.common.template_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_setup_heat_environments', autospec=True)
    @mock.patch('tripleo_common.image.kolla_builder.'
                'container_images_prepare_multi')
    def test_deploy_tripleo_heat_templates_redir(self,
                                                 mock_cipm,
                                                 mock_setup_heat_envs,
                                                 mock_hc_templ_parse,
                                                 mock_hc_env_parse,
                                                 mock_hc_get_templ_cont,
                                                 mock_hc_process):

        with tempfile.NamedTemporaryFile(delete=False) as roles_file:
            self.addCleanup(os.unlink, roles_file.name)

        mock_cipm.return_value = {}

        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot',
                                         '--roles-file', roles_file.name], [])

        mock_setup_heat_envs.return_value = [
            './inside.yaml', '/tmp/thtroot/abs.yaml',
            '/tmp/thtroot/puppet/foo.yaml',
            '/tmp/thtroot/environments/myenv.yaml',
            '/tmp/thtroot42/notouch.yaml',
            '../outside.yaml']

        self.cmd._deploy_tripleo_heat_templates(self.orc, parsed_args)

        mock_hc_process.assert_has_calls([
            mock.call(env_path='./inside.yaml'),
            mock.call(env_path='/twd/templates/abs.yaml'),
            mock.call(env_path='/twd/templates/puppet/foo.yaml'),
            mock.call(env_path='/twd/templates/environments/myenv.yaml'),
            mock.call(env_path='/tmp/thtroot42/notouch.yaml'),
            mock.call(env_path='../outside.yaml')])

    @mock.patch('tripleoclient.utils.rel_or_abs_path')
    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.environment_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('heatclient.common.template_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_setup_heat_environments', autospec=True)
    @mock.patch('yaml.safe_dump', autospec=True)
    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tempfile.NamedTemporaryFile', autospec=True)
    @mock.patch('tripleo_common.image.kolla_builder.'
                'container_images_prepare_multi')
    def test_deploy_tripleo_heat_templates_rewrite(self,
                                                   mock_cipm,
                                                   mock_temp, mock_open,
                                                   mock_yaml_load,
                                                   mock_yaml_dump,
                                                   mock_setup_heat_envs,
                                                   mock_hc_templ_parse,
                                                   mock_hc_env_parse,
                                                   mock_hc_get_templ_cont,
                                                   mock_hc_process,
                                                   mock_norm_path):
        def hc_process(*args, **kwargs):
            if 'abs.yaml' in kwargs['env_path']:
                raise hc_exc.CommandError
            else:
                return ({}, {})

        mock_cipm.return_value = {}

        mock_hc_process.side_effect = hc_process

        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', '/tmp/thtroot'], [])

        rewritten_env = {'resource_registry': {
            'OS::Foo::Bar': '/twd/outside.yaml',
            'OS::Foo::Baz': '/twd/templates/inside.yaml',
            'OS::Foo::Qux': '/twd/templates/abs.yaml',
            'OS::Foo::Quux': '/tmp/thtroot42/notouch.yaml',
            'OS::Foo::Corge': '/twd/templates/puppet/foo.yaml'
            }
        }
        myenv = {'resource_registry': {
            'OS::Foo::Bar': '../outside.yaml',
            'OS::Foo::Baz': './inside.yaml',
            'OS::Foo::Qux': '/tmp/thtroot/abs.yaml',
            'OS::Foo::Quux': '/tmp/thtroot42/notouch.yaml',
            'OS::Foo::Corge': '/tmp/thtroot/puppet/foo.yaml'
            }
        }
        mock_yaml_load.return_value = myenv

        mock_setup_heat_envs.return_value = [
            './inside.yaml', '/tmp/thtroot/abs.yaml',
            '/tmp/thtroot/puppet/foo.yaml',
            '/tmp/thtroot/environments/myenv.yaml',
            '../outside.yaml']

        self.cmd._deploy_tripleo_heat_templates(self.orc, parsed_args)

        mock_yaml_dump.assert_has_calls([mock.call(rewritten_env,
                                        default_flow_style=False)])

    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_populate_templates_dir')
    @mock.patch('tripleoclient.utils.fetch_roles_file')
    @mock.patch('tripleoclient.utils.rel_or_abs_path')
    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.environment_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('heatclient.common.template_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_setup_heat_environments', autospec=True)
    @mock.patch('yaml.safe_dump', autospec=True)
    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tempfile.NamedTemporaryFile', autospec=True)
    @mock.patch('tripleo_common.image.kolla_builder.'
                'container_images_prepare_multi')
    def test_deploy_tripleo_heat_templates_remove(self,
                                                  mock_cipm,
                                                  mock_temp, mock_open,
                                                  mock_yaml_load,
                                                  mock_yaml_dump,
                                                  mock_setup_heat_envs,
                                                  mock_hc_templ_parse,
                                                  mock_hc_env_parse,
                                                  mock_hc_get_templ_cont,
                                                  mock_hc_process,
                                                  mock_norm_path,
                                                  mock_fetch_roles,
                                                  mock_populate):
        def hc_process(*args, **kwargs):
            if 'myenv.yaml' in kwargs['env_path']:
                env = {
                    'resource_registry': {
                        'OS::TripleO::Services::Foo': 'OS::Heat::None'}}
                return ({}, env)
            else:
                return ({}, {})

        mock_fetch_roles.return_value = [
            {'name': 'Bar', 'ServicesDefault': [
                'OS::TripleO::Services::Foo', 'OS::TripleO::Services::Bar']},
            {'name': 'Foo', 'tags': ['primary']}
        ]

        def set_tht(templates, stack_name):
            self.cmd.tht_render = "tht_from"

        mock_populate.side_effect = set_tht

        mock_cipm.return_value = {}

        mock_hc_process.side_effect = hc_process

        parsed_args = self.check_parser(self.cmd,
                                        ['--templates', '/tmp/thtroot'], [])

        rewritten_role = [
            {'name': 'Bar', 'ServicesDefault': ['OS::TripleO::Services::Bar']},
            {'name': 'Foo', 'tags': ['primary']}
        ]
        myenv = {'resource_registry': {
            'OS::Foo::Bar': '../outside.yaml',
            'OS::Foo::Baz': './inside.yaml',
            'OS::Foo::Qux': '/tmp/thtroot/abs.yaml',
            'OS::Foo::Quux': '/tmp/thtroot42/notouch.yaml',
            'OS::Foo::Corge': '/tmp/thtroot/puppet/foo.yaml'
            }
        }
        mock_yaml_load.return_value = myenv

        mock_setup_heat_envs.return_value = [
            './inside.yaml', '/tmp/thtroot/abs.yaml',
            '/tmp/thtroot/puppet/foo.yaml',
            '/tmp/thtroot/environments/myenv.yaml',
            '../outside.yaml']

        self.cmd._deploy_tripleo_heat_templates(self.orc, parsed_args)

        mock_yaml_dump.assert_has_calls([mock.call(rewritten_role)])

    @mock.patch('shutil.copy')
    @mock.patch('os.path.exists', return_value=False)
    def test_normalize_user_templates(self, mock_exists, mock_copy):
        user_tht_root = '/userroot'
        tht_root = '/thtroot'
        env_files = [
            '/home/basic.yaml',
            '/home/dir/dir.yaml',
            'home/relative.yaml',
            'file.yaml',
            '~/tilde.yaml',
            '../../../dots.yaml',
            '/userroot/template.yaml',
            '/userroot/tht/tht.yaml',
        ]
        expected = [
            '/thtroot/basic.yaml',
            '/thtroot/dir.yaml',
            '/thtroot/relative.yaml',
            '/thtroot/file.yaml',
            '/thtroot/tilde.yaml',
            '/thtroot/dots.yaml',
            '/thtroot/template.yaml',
            '/thtroot/tht/tht.yaml'
        ]
        results = self.cmd._normalize_user_templates(user_tht_root,
                                                     tht_root,
                                                     env_files)

        self.assertEqual(expected, results)
        self.assertEqual(mock_copy.call_count, 6)

    @mock.patch('os.path.exists', return_value=True)
    def test_normalize_user_templates_exists(self, mock_exists):
        user_tht_root = '/userroot'
        tht_root = '/thtroot'
        env_files = ['/home/basic.yaml']
        self.assertRaises(exceptions.DeploymentError,
                          self.cmd._normalize_user_templates,
                          user_tht_root,
                          tht_root,
                          env_files)

    @mock.patch('os.path.exists', return_value=True)
    def test_normalize_user_templates_trailing_slash(self, mock_exists):
        user_tht_root = '/userroot/'
        tht_root = '/thtroot'
        env_files = ['/userroot/basic.yaml']
        expected = ['/thtroot/basic.yaml']
        results = self.cmd._normalize_user_templates(user_tht_root,
                                                     tht_root,
                                                     env_files)
        self.assertEqual(expected, results)

    @mock.patch('time.time', return_value=123)
    @mock.patch('yaml.safe_load', return_value={}, autospec=True)
    @mock.patch('yaml.safe_dump', autospec=True)
    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_process_hieradata_overrides', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_update_passwords_env', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_normalize_user_templates', return_value=[], autospec=True)
    @mock.patch('tripleoclient.utils.rel_or_abs_path', return_value={},
                autospec=True)
    @mock.patch('tripleoclient.utils.run_command_and_log', return_value=0,
                autospec=True)
    def test_setup_heat_environments_dropin(
            self, mock_run, mock_paths, mock_norm, mock_update_pass_env,
            mock_process_hiera, mock_open, mock_os, mock_yaml_dump,
            mock_yaml_load, mock_time):

        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', 'tht_from',
                                         '--output-dir', 'tht_to'], [])
        dropin = 'tht_from/standalone-stack-vstate-dropin.yaml'
        self.cmd.output_dir = 'tht_to'
        self.cmd.tht_render = 'tht_from'
        self.cmd.stack_action = 'UPDATE'
        environment = self.cmd._setup_heat_environments(
            parsed_args.roles_file, parsed_args.networks_file, parsed_args)

        self.assertIn(dropin, environment)
        mock_open.assert_has_calls([mock.call(dropin, 'w')])

        # unpack the dump yaml calls to verify if the produced stack update
        # dropin matches our expectations
        found_dropin = False
        found_identifier = False
        for call in mock_yaml_dump.call_args_list:
            args, kwargs = call
            for a in args:
                if isinstance(a, mock.mock.MagicMock):
                    continue
                if a.get('parameter_defaults', {}).get('StackAction', None):
                    self.assertTrue(
                        a['parameter_defaults']['StackAction'] == 'UPDATE')
                    found_dropin = True
                if a.get('parameter_defaults', {}).get('DeployIdentifier',
                                                       None):
                    self.assertTrue(
                        a['parameter_defaults']['DeployIdentifier'] == 123)
                    found_identifier = True
        self.assertTrue(found_dropin)
        self.assertTrue(found_identifier)

    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('tripleoclient.utils.'
                'process_multiple_environments', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_process_hieradata_overrides', return_value='hiera_or.yaml',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_update_passwords_env', autospec=True)
    @mock.patch('tripleoclient.utils.'
                'run_command_and_log', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_get_primary_role_name', autospec=True)
    def test_setup_heat_environments_default_plan_env(
            self, mock_prim, mock_run, mock_update_pass_env,
            mock_process_hiera, mock_process_multiple_environments,
            mock_hc_get_templ_cont, mock_hc_process):

        tmpdir = self.useFixture(fixtures.TempDir()).path
        tht_from = os.path.join(tmpdir, 'tht-from')
        os.mkdir(tht_from)
        plan_env_path = os.path.join(tht_from, 'plan-environment.yaml')
        with open(plan_env_path, mode='w') as plan_file:
            yaml.dump({'environments': [{'path': 'env.yaml'}]}, plan_file)
        self.assertTrue(os.path.exists(plan_env_path))
        self._setup_heat_environments(tmpdir, tht_from, plan_env_path,
                                      mock_update_pass_env, mock_run)

    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('tripleoclient.utils.'
                'process_multiple_environments', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_process_hieradata_overrides', return_value='hiera_or.yaml',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_update_passwords_env', autospec=True)
    @mock.patch('tripleoclient.utils.'
                'run_command_and_log', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_get_primary_role_name', autospec=True)
    def test_setup_heat_environments_non_default_plan_env(
            self, mock_prim, mock_run, mock_update_pass_env,
            mock_process_hiera, mock_process_multiple_environments,
            mock_hc_get_templ_cont, mock_hc_process):

        tmpdir = self.useFixture(fixtures.TempDir()).path
        tht_from = os.path.join(tmpdir, 'tht-from')
        os.mkdir(tht_from)
        default_plan_env_path = os.path.join(tht_from, 'plan-environment.yaml')
        with open(default_plan_env_path, mode='w') as plan_file:
            yaml.dump({'environments': [{'path': 'env.yaml'}]}, plan_file)
        plan_env_path = os.path.join(tmpdir, 'plan-environment.yaml')
        with open(plan_env_path, mode='w') as plan_file:
            yaml.dump({'environments': [{'path': 'notenv.yaml'}]}, plan_file)
        self.assertTrue(os.path.exists(plan_env_path))
        with open(os.path.join(tht_from, 'notenv.yaml'),
                  mode='w') as env_file:
            yaml.dump({}, env_file)
        cmd_extra = ['-p', plan_env_path]
        self._setup_heat_environments(tmpdir, tht_from, plan_env_path,
                                      mock_update_pass_env, mock_run,
                                      cmd_extra, 'notenv.yaml')

    def _setup_heat_environments(self, tmpdir, tht_from, plan_env_path,
                                 mock_update_pass_env, mock_run,
                                 extra_cmd=None, plan_env_env=None):
        cmd_extra = extra_cmd or []
        plan_env_env_name = plan_env_env or 'env.yaml'

        tht_outside = os.path.join(tmpdir, 'tht-outside')
        os.mkdir(tht_outside)
        tht_to = os.path.join(tmpdir, 'tht-to')
        os.mkdir(tht_to)
        with open(os.path.join(tht_from, 'env.yaml'),
                  mode='w') as env_file:
            yaml.dump({}, env_file)
        with open(os.path.join(tht_from, 'foo.yaml'),
                  mode='w') as env_file:
            yaml.dump({}, env_file)
        with open(os.path.join(tht_outside, 'outside.yaml'),
                  mode='w') as env_file:
            yaml.dump({}, env_file)

        tht_render = os.path.join(tht_to, 'tripleo-heat-installer-templates')
        mock_update_pass_env.return_value = os.path.join(
            tht_render, 'passwords.yaml')
        mock_run.return_value = 0
        original_abs = os.path.abspath

        # Stub abspath for default plan and envs to return the tht_render base
        def abs_path_stub(*args, **kwargs):
            if 'plan-environment.yaml' in args:
                return plan_env_path
            elif 'notenv.yaml' in args:
                return os.path.join(tht_render, 'notenv.yaml')
            elif 'env.yaml' in args:
                return os.path.join(tht_render, 'env.yaml')
            else:
                return original_abs(*args, **kwargs)

        # logic handled in _standalone_deploy
        self.cmd.output_dir = tht_to
        # Note we don't create tht_render as _populate_templates_dir creates it
        self.cmd.tht_render = tht_render
        self.cmd._populate_templates_dir(tht_from)

        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1/8',
                                         '--templates', tht_from,
                                         '--output-dir', tht_to,
                                         '--hieradata-override',
                                         'legacy.yaml',
                                         '-e',
                                         os.path.join(tht_from, 'foo.yaml'),
                                         '-e',
                                         os.path.join(tht_outside,
                                                      'outside.yaml'),
                                         ] + cmd_extra, [])
        expected_env = [
            os.path.join(tht_render, plan_env_env_name),
            os.path.join(tht_render, 'passwords.yaml'),
            os.path.join(tht_render,
                         'environments/deployed-server-noop-ctlplane.yaml'),
            os.path.join(tht_render,
                         'tripleoclient-hosts-portmaps.yaml'),
            'hiera_or.yaml',
            os.path.join(tht_render, 'standalone-stack-vstate-dropin.yaml'),
            os.path.join(tht_render, 'foo.yaml'),
            os.path.join(tht_render, 'outside.yaml')]

        with mock.patch('os.path.abspath', side_effect=abs_path_stub):
            with mock.patch('os.path.isfile'):
                environment = self.cmd._setup_heat_environments(
                    parsed_args.roles_file, parsed_args.networks_file,
                    parsed_args)

                self.assertEqual(expected_env, environment)

    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs', autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.TripleoInventory',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_launch_heat', autospec=True)
    @mock.patch('tripleo_common.utils.config.Config',
                autospec=True)
    @mock.patch('os.path.join', return_value='/twd/inventory.yaml')
    def test_download_ansible_playbooks(self, mock_join, mock_stack_config,
                                        mock_launch_heat, mock_importInv,
                                        createdir_mock):

        fake_output_dir = '/twd'
        extra_vars = {'Undercloud': {
            'ansible_connection': 'local',
            'ansible_python_interpreter': sys.executable}}
        mock_inventory = mock.Mock()
        mock_importInv.return_value = mock_inventory
        with mock.patch('sys.stdout', autospec=True) as mock_stdout:
            self.cmd.output_dir = fake_output_dir
            self.cmd._download_ansible_playbooks(mock_launch_heat,
                                                 'undercloud',
                                                 'Undercloud')
            self.assertEqual(mock_stdout.flush.call_count, 2)
        mock_inventory.write_static_inventory.assert_called_once_with(
            fake_output_dir + '/inventory.yaml', extra_vars)

    @mock.patch('tripleo_common.image.kolla_builder.'
                'container_images_prepare_multi')
    def test_prepare_container_images(self, mock_cipm):
        env = {'parameter_defaults': {}}
        mock_cipm.return_value = {'FooImage': 'foo/bar:baz'}

        self.cmd._prepare_container_images(env, [{'name': 'Compute'}])

        mock_cipm.assert_called_once_with(
            env,
            [{'name': 'Compute'}],
            dry_run=True,
        )
        self.assertEqual(
            {
                'parameter_defaults': {
                    'FooImage': 'foo/bar:baz'
                }
            },
            env
        )

    @mock.patch.object(
        ansible_runner.runner_config,
        'RunnerConfig',
        return_value=fakes.FakeRunnerConfig()
    )
    @mock.patch.object(
        ansible_runner.Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return()
    )
    @mock.patch('os.path.exists')
    @mock.patch('os.chdir')
    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_stack_outputs')
    @mock.patch('tripleo_common.actions.ansible.'
                'write_default_ansible_cfg')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.chmod')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('subprocess.check_call', autospec=True)
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('getpass.getuser', return_value='stack')
    @mock.patch('os.mkdir')
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_populate_templates_dir')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_install_artifact', return_value='/tmp/foo.tar.bzip2')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_cleanup_working_dirs')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    @mock.patch('tripleoclient.utils.wait_api_port_ready',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_deploy_tripleo_heat_templates', autospec=True,
                return_value='undercloud, 0')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_ansible_playbooks', autospec=True,
                return_value='/foo')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_launch_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_kill_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_configure_puppet')
    @mock.patch('os.geteuid', return_value=0)
    @mock.patch('os.environ', return_value='CREATE_COMPLETE')
    @mock.patch('tripleoclient.utils.wait_for_stack_ready', return_value=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_set_default_plan')
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_take_action_standalone(self, mock_dump_artifact,
                                    mock_def_plan, mock_poll,
                                    mock_environ, mock_geteuid, mock_puppet,
                                    mock_killheat, mock_launchheat,
                                    mock_download, mock_tht,
                                    mock_wait_for_port, mock_createdirs,
                                    mock_cleanupdirs, mock_tarball,
                                    mock_templates_dir, mock_open, mock_os,
                                    mock_user, mock_cc, mock_chmod, mock_ac,
                                    mock_outputs, mock_copy, mock_cmdline,
                                    mock_chdir, mock_file_exists, mock_run,
                                    mock_run_prepare):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--standalone-role', 'Undercloud',
                                         # TODO(cjeanner) drop once we have
                                         # proper oslo.privsep
                                         '--deployment-user', 'stack',
                                         '-e', '/tmp/thtroot/puppet/foo.yaml',
                                         '-e', '/tmp/thtroot//docker/bar.yaml',
                                         '-e', '/tmp/thtroot42/notouch.yaml',
                                         '-e', '~/custom.yaml',
                                         '-e', 'something.yaml',
                                         '-e', '../../../outside.yaml'], [])

        mock_file_exists.return_value = True
        fake_orchestration = mock_launchheat(parsed_args)
        self.cmd.take_action(parsed_args)
        mock_createdirs.assert_called_once()
        mock_puppet.assert_called_once()
        mock_launchheat.assert_called_with(parsed_args)
        mock_tht.assert_called_once_with(self.cmd, fake_orchestration,
                                         parsed_args)
        mock_download.assert_called_with(self.cmd, fake_orchestration,
                                         'undercloud', 'Undercloud',
                                         sys.executable)
        mock_tarball.assert_called_once()
        mock_cleanupdirs.assert_called_once()
        self.assertEqual(mock_killheat.call_count, 2)

    @mock.patch.object(
        ansible_runner.runner_config,
        'RunnerConfig',
        return_value=fakes.FakeRunnerConfig()
    )
    @mock.patch.object(
        ansible_runner.Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return(1)
    )
    @mock.patch('os.path.exists')
    @mock.patch('os.chdir')
    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_stack_outputs')
    @mock.patch('tripleo_common.actions.ansible.'
                'write_default_ansible_cfg')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.chmod')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('subprocess.check_call', autospec=True)
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('getpass.getuser', return_value='stack')
    @mock.patch('os.mkdir')
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_populate_templates_dir')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_install_artifact', return_value='/tmp/foo.tar.bzip2')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_cleanup_working_dirs')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    @mock.patch('tripleoclient.utils.wait_api_port_ready',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_deploy_tripleo_heat_templates', autospec=True,
                return_value='undercloud, 0')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_ansible_playbooks', autospec=True,
                return_value='/foo')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_launch_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_kill_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_configure_puppet')
    @mock.patch('os.geteuid', return_value=0)
    @mock.patch('os.environ', return_value='CREATE_COMPLETE')
    @mock.patch('tripleoclient.utils.wait_for_stack_ready', return_value=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_set_default_plan')
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_take_action_ansible_err(self, mock_dump_artifact,
                                     mock_def_plan, mock_poll,
                                     mock_environ, mock_geteuid, mock_puppet,
                                     mock_killheat, mock_launchheat,
                                     mock_download, mock_tht,
                                     mock_wait_for_port, mock_createdirs,
                                     mock_cleanupdirs, mock_tarball,
                                     mock_templates_dir, mock_open, mock_os,
                                     mock_user, mock_cc, mock_chmod, mock_ac,
                                     mock_outputs, mock_copy, mock_cmdline,
                                     mock_chdir, mock_file_exists, mock_run,
                                     mock_run_prepare):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--standalone-role', 'Undercloud',
                                         # TODO(cjeanner) drop once we have
                                         # proper oslo.privsep
                                         '--deployment-user', 'stack',
                                         '-e', '/tmp/thtroot/puppet/foo.yaml',
                                         '-e', '/tmp/thtroot//docker/bar.yaml',
                                         '-e', '/tmp/thtroot42/notouch.yaml',
                                         '-e', '~/custom.yaml',
                                         '-e', 'something.yaml',
                                         '-e', '../../../outside.yaml'], [])

        mock_file_exists.return_value = True
        fake_orchestration = mock_launchheat(parsed_args)
        self.assertRaises(exceptions.DeploymentError,
                          self.cmd.take_action, parsed_args)
        mock_createdirs.assert_called_once()
        mock_puppet.assert_called_once()
        mock_launchheat.assert_called_with(parsed_args)
        mock_tht.assert_called_once_with(self.cmd, fake_orchestration,
                                         parsed_args)
        mock_download.assert_called_with(self.cmd, fake_orchestration,
                                         'undercloud', 'Undercloud',
                                         sys.executable)
        mock_tarball.assert_called_once()
        mock_cleanupdirs.assert_called_once()
        self.assertEqual(mock_killheat.call_count, 2)

    @mock.patch('os.chdir')
    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_stack_outputs')
    @mock.patch('tripleo_common.actions.ansible.'
                'write_default_ansible_cfg')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.chmod')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('subprocess.check_call', autospec=True)
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('getpass.getuser', return_value='stack')
    @mock.patch('os.mkdir')
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_populate_templates_dir')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_install_artifact', return_value='/tmp/foo.tar.bzip2')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_cleanup_working_dirs')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    @mock.patch('tripleoclient.utils.wait_api_port_ready')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_deploy_tripleo_heat_templates', autospec=True,
                return_value='undercloud, 0')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_ansible_playbooks', autospec=True,
                return_value='/foo')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_launch_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_kill_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_configure_puppet')
    @mock.patch('os.geteuid', return_value=0)
    @mock.patch('os.environ', return_value='CREATE_COMPLETE')
    @mock.patch('tripleoclient.utils.wait_for_stack_ready', return_value=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_set_default_plan')
    def test_take_action_other_err(self,
                                   mock_def_plan, mock_poll,
                                   mock_environ, mock_geteuid, mock_puppet,
                                   mock_killheat, mock_launchheat,
                                   mock_download, mock_tht,
                                   mock_wait_for_port, mock_createdirs,
                                   mock_cleanupdirs, mock_tarball,
                                   mock_templates_dir, mock_open, mock_os,
                                   mock_user, mock_cc, mock_chmod, mock_ac,
                                   mock_outputs, mock_copy, mock_cmdline,
                                   mock_chdir):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--standalone-role', 'Undercloud',
                                         # TODO(cjeanner) drop once we have
                                         # proper oslo.privsep
                                         '--deployment-user', 'stack',
                                         '-e', '/tmp/thtroot/puppet/foo.yaml',
                                         '-e', '/tmp/thtroot//docker/bar.yaml',
                                         '-e', '/tmp/thtroot42/notouch.yaml',
                                         '-e', '~/custom.yaml',
                                         '-e', 'something.yaml',
                                         '-e', '../../../outside.yaml'], [])

        mock_wait_for_port.side_effect = exceptions.DeploymentError
        self.assertRaises(exceptions.DeploymentError,
                          self.cmd.take_action, parsed_args)
        mock_createdirs.assert_called_once()
        mock_puppet.assert_called_once()
        mock_launchheat.assert_called_with(parsed_args)
        mock_tht.assert_not_called()
        mock_download.assert_not_called()
        mock_tarball.assert_called_once()
        mock_cleanupdirs.assert_called_once()
        self.assertEqual(mock_killheat.call_count, 1)

    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    def test_take_action(self, mock_copy, mock_cmdline):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my'], [])
        self.assertRaises(exceptions.DeploymentError,
                          self.cmd.take_action, parsed_args)
        mock_copy.assert_called_once()

    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy._standalone_deploy')
    def test_take_action_failure(self, mock_deploy, mock_copy, mock_cmdline):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my'], [])
        mock_deploy.side_effect = exceptions.DeploymentError
        self.assertRaises(exceptions.DeploymentError,
                          self.cmd.take_action, parsed_args)
        mock_copy.assert_called_once()

    @mock.patch('os.path.isfile', return_value=False)
    def test_set_stack_action_default_create(self, mock_isfile):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my'], [])
        self.cmd._set_stack_action(parsed_args)
        self.assertEqual('CREATE', self.cmd.stack_action)

    @mock.patch('os.path.isfile', return_value=True)
    def test_set_stack_action_default_update(self, mock_isfile):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my'], [])
        self.cmd._set_stack_action(parsed_args)
        self.assertEqual('UPDATE', self.cmd.stack_action)

    @mock.patch('os.path.isfile', return_value=False)
    def test_set_stack_action_force_update(self, mock_isfile):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--force-stack-update'], [])
        self.cmd._set_stack_action(parsed_args)
        self.assertEqual('UPDATE', self.cmd.stack_action)

    @mock.patch('os.path.isfile', return_value=True)
    def test_set_stack_action_force_create(self, mock_isfile):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--force-stack-create'], [])
        self.cmd._set_stack_action(parsed_args)
        self.assertEqual('CREATE', self.cmd.stack_action)

    @mock.patch('os.path.isfile', return_value=True)
    def test_set_stack_action_mutually_exclusive(self, mock_isfile):
        self.assertRaises(
            SystemExit,
            self.check_parser,
            self.cmd,
            ['--local-ip', '127.0.0.1',
             '--templates', '/tmp/thtroot',
             '--stack', 'undercloud',
             '--output-dir', '/my',
             '--force-stack-create',
             '--force-stack-update'], [])

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.chdir')
    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_stack_outputs')
    @mock.patch('tripleo_common.actions.ansible.'
                'write_default_ansible_cfg')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.chmod')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.mkdir')
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_populate_templates_dir')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_install_artifact', return_value='/tmp/foo.tar.bzip2')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_cleanup_working_dirs')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    @mock.patch('tripleoclient.utils.wait_api_port_ready',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_deploy_tripleo_heat_templates', autospec=True,
                return_value='undercloud, 0')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_ansible_playbooks', autospec=True,
                return_value='/foo')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_launch_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_kill_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_configure_puppet')
    @mock.patch('os.geteuid', return_value=0)
    @mock.patch('os.environ', return_value='CREATE_COMPLETE')
    @mock.patch('tripleoclient.utils.wait_for_stack_ready', return_value=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_set_default_plan')
    def test_standalone_deploy_rc_output_only(
            self,
            mock_def_plan, mock_poll,
            mock_environ, mock_geteuid, mock_puppet,
            mock_killheat, mock_launchheat,
            mock_download, mock_tht,
            mock_wait_for_port, mock_createdirs,
            mock_cleanupdirs, mock_tarball,
            mock_templates_dir, mock_open, mock_os,
            mock_chmod, mock_ac,
            mock_outputs, mock_copy, mock_cmdline,
            mock_chdir, mock_file_exists):
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--output-only'], [])

        rc = self.cmd.take_action(parsed_args)
        self.assertEqual(None, rc)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.chdir')
    @mock.patch('tripleoclient.utils.reset_cmdline')
    @mock.patch('tripleoclient.utils.copy_clouds_yaml')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_stack_outputs')
    @mock.patch('tripleo_common.actions.ansible.'
                'write_default_ansible_cfg')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.chmod')
    # TODO(cjeanner) drop once we have proper oslo.privsep
    @mock.patch('os.mkdir')
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_populate_templates_dir')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_install_artifact', return_value='/tmp/foo.tar.bzip2')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_cleanup_working_dirs')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_create_working_dirs')
    @mock.patch('tripleoclient.utils.wait_api_port_ready',
                autospec=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_deploy_tripleo_heat_templates', autospec=True,
                return_value='undercloud, 0')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_download_ansible_playbooks', autospec=True,
                return_value='/foo')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_launch_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_kill_heat')
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_configure_puppet')
    @mock.patch('os.geteuid', return_value=0)
    @mock.patch('os.environ', return_value='CREATE_COMPLETE')
    @mock.patch('tripleoclient.utils.wait_for_stack_ready', return_value=True)
    @mock.patch('tripleoclient.v1.tripleo_deploy.Deploy.'
                '_set_default_plan')
    def test_standalone_deploy_transport(
            self,
            mock_def_plan, mock_poll,
            mock_environ, mock_geteuid, mock_puppet,
            mock_killheat, mock_launchheat,
            mock_download, mock_tht,
            mock_wait_for_port, mock_createdirs,
            mock_cleanupdirs, mock_tarball,
            mock_templates_dir, mock_open, mock_os,
            mock_chmod, mock_ac,
            mock_outputs, mock_copy, mock_cmdline,
            mock_chdir, mock_file_exists):

        # Test default transport "local"
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--output-only'], [])

        self.cmd.take_action(parsed_args)
        self.assertEqual(1, mock_ac.call_count)
        self.assertEqual(
            "local", mock_ac.call_args_list[0].kwargs["transport"])

        # Test transport "ssh"
        mock_ac.reset_mock()
        parsed_args = self.check_parser(self.cmd,
                                        ['--local-ip', '127.0.0.1',
                                         '--templates', '/tmp/thtroot',
                                         '--stack', 'undercloud',
                                         '--output-dir', '/my',
                                         '--output-only',
                                         '--transport', 'ssh'], [])

        self.cmd.take_action(parsed_args)
        self.assertEqual(1, mock_ac.call_count)
        self.assertEqual("ssh", mock_ac.call_args_list[0].kwargs["transport"])



#   Copyright 2018 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from oslo_config import cfg
from oslo_log import log as logging

from tripleoclient import constants
from tripleoclient.exceptions import UndercloudUpgradeNotConfirmed
from tripleoclient import utils
from tripleoclient.v1.tripleo_deploy import Deploy

CONF = cfg.CONF


class Upgrade(Deploy):
    """Upgrade TripleO"""

    log = logging.getLogger(__name__ + ".Upgrade")

    def take_action(self, parsed_args):
        logging.register_options(CONF)
        logging.setup(CONF, '')
        self.log.debug("take_action(%s)" % parsed_args)
        if (not parsed_args.yes
                and not utils.prompt_user_for_confirmation(
                    constants.UPGRADE_PROMPT, self.log)):
            raise UndercloudUpgradeNotConfirmed(constants.UPGRADE_NO)

        parsed_args.standalone = True
        parsed_args.upgrade = True
        super(Upgrade, self).take_action(parsed_args)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import os
import yaml

from tripleo_common.actions import plan
from tripleo_common.utils import plan as plan_utils
from tripleo_common.utils import swift as swiftutils
from tripleo_common.utils import template as temputils

from tripleoclient import constants
from tripleoclient import exceptions
from tripleoclient import utils

LOG = logging.getLogger(__name__)
# Plan management workflows should generally be quick. However, the creation
# of the default plan in instack has demonstrated that sometimes it can take
# several minutes. The previous timeout of 6 minutes from Instack does not
# seem to be sufficient anymore. Bumping this to 20 minutes. It doesn't mean
# that it will take 20 minutes, but just that the listen for completion will
# timeout after 20 minutes. If it takes longer than that, something is really
# wrong.
_WORKFLOW_TIMEOUT = 20 * 60  # 20 minutes * 60 seconds


def _upload_templates(swift_client, container_name, tht_root, roles_file=None,
                      plan_env_file=None, networks_file=None):
    """tarball up a given directory and upload it to Swift to be extracted"""

    temputils.upload_templates_as_tarball(
        swift=swift_client,
        dir_to_upload=tht_root,
        container=container_name
    )

    # Optional override of the roles_data.yaml file
    if roles_file:
        _upload_file(swift_client, container_name,
                     constants.OVERCLOUD_ROLES_FILE,
                     utils.rel_or_abs_path(roles_file, tht_root))

    # Optional override of the network_data.yaml file
    if networks_file:
        _upload_file(swift_client, container_name,
                     constants.OVERCLOUD_NETWORKS_FILE, networks_file)

    # Optional override of the plan-environment.yaml file
    if plan_env_file:
        # TODO(jpalanis): Instead of overriding default file,
        # merging the user override plan-environment with default
        # plan-environment file will avoid explict merging issues.
        _upload_file(swift_client, container_name,
                     constants.PLAN_ENVIRONMENT, plan_env_file)


def create_deployment_plan(container, generate_passwords,
                           use_default_templates=False, source_url=None,
                           verbosity_level=0, plan_env_file=None,
                           disable_image_params_prepare=False):
    """Create a deployment plan.

    :param container: Container name to push to.
    :type container: String

    :param generate_passwords: Whether to generate password
    :type generate_passwords: Boolean

    :param use_default_templates: Whether to use default template files
    :type use_default_templates: Boolean

    :param source_url: The url of a git repository containing the Heat
                       templates to deploy
    :type source_url: String

    :param verbosity_level: Verbosity level used in playbook execution
    :type verbosity_level: Integer

    :param plan_env_file: Path to plan environment file
    :type plan_env_file: String

    :param disable_image_params_prepare: Disable container params prepare task
    :type disable_image_params_prepare: Boolean
    """

    extra_vars = {
        "container": container,
        "generate_passwords": generate_passwords,
        "use_default_templates": use_default_templates,
        "disable_image_params_prepare": disable_image_params_prepare
    }

    if source_url:
        extra_vars['source_url'] = source_url

    if plan_env_file:
        extra_vars['plan_environment'] = plan_env_file

    with utils.TempDirs() as tmp:
        utils.run_ansible_playbook(
            "cli-create-deployment-plan.yaml",
            'undercloud,',
            workdir=tmp,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars=extra_vars,
            verbosity=verbosity_level
        )

    print("Success.")


def delete_deployment_plan(clients, container):
    """Delete a deployment plan.

    :param clients: Application client object.
    :type clients: Object

    :param container: Container name to pull from.
    :type container: String
    """

    context = clients.tripleoclient.create_mistral_context()
    result = plan.DeletePlanAction(container=container).run(context=context)
    # The action returns None if there are no errors.
    if result:
        raise RuntimeError(result)


def update_deployment_plan(clients, container, generate_passwords,
                           verbosity_level=0,
                           disable_image_params_prepare=False):
    """Update a deployment plan.

    :param clients: Application client object.
    :type clients: Object

    :param container: Container name to pull from.
    :type container: String

    :param generate_passwords: Whether to generate password
    :type generate_passwords: Boolean

    :param verbosity_level: Verbosity level used in playbook execution
    :type verbosity_level: Integer

    :param disable_image_params_prepare: Disable container params prepare task
    :type disable_image_params_prepare: Boolean
    """

    with utils.TempDirs() as tmp:
        utils.run_ansible_playbook(
            "cli-update-deployment-plan.yaml",
            'undercloud,',
            workdir=tmp,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "container": container,
                "generate_passwords": generate_passwords,
                "disable_image_params_prepare": disable_image_params_prepare
            },
            verbosity=verbosity_level
        )

    print("Success.")


def list_deployment_plans(clients):
    mistral_context = clients.tripleoclient.create_mistral_context()
    return plan.ListPlansAction().run(mistral_context)


def create_plan_from_templates(clients, name, tht_root, roles_file=None,
                               generate_passwords=True, plan_env_file=None,
                               networks_file=None, verbosity_level=0,
                               disable_image_params_prepare=False):

    with utils.TempDirs() as tmp:
        utils.run_ansible_playbook(
            "cli-undercloud-local-artifacts.yaml",
            'undercloud,',
            workdir=tmp,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "stack": name,
            },
            verbosity=verbosity_level
        )

    swift_client = clients.tripleoclient.object_store

    print("Creating Swift container to store the plan")
    plan_utils.create_plan_container(swift_client, name)

    print("Creating plan from template files in: {}".format(tht_root))
    _upload_templates(swift_client, name, tht_root,
                      utils.rel_or_abs_path(roles_file, tht_root),
                      plan_env_file, networks_file)

    try:
        create_deployment_plan(
            container=name,
            generate_passwords=generate_passwords,
            plan_env_file=plan_env_file,
            verbosity_level=verbosity_level,
            disable_image_params_prepare=disable_image_params_prepare)
    except exceptions.WorkflowServiceError:
        swiftutils.delete_container(swift_client, name)
        raise


def update_plan_from_templates(clients, name, tht_root, roles_file=None,
                               generate_passwords=True, plan_env_file=None,
                               networks_file=None, keep_env=False,
                               verbosity_level=1,
                               disable_image_params_prepare=False):

    with utils.TempDirs() as tmp:
        utils.run_ansible_playbook(
            "cli-undercloud-local-artifacts.yaml",
            'undercloud,',
            workdir=tmp,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "stack": name,
            },
            verbosity=verbosity_level
        )

    swift_client = clients.tripleoclient.object_store
    passwords = None
    keep_file_contents = {}
    roles_file = utils.rel_or_abs_path(roles_file, tht_root)

    if keep_env:
        # Dict items are (remote_name, local_name). local_name may be
        # None in which case we only try to load from Swift (remote).
        keep_map = {
            constants.PLAN_ENVIRONMENT: plan_env_file,
            constants.USER_ENVIRONMENT: None,
            constants.OVERCLOUD_ROLES_FILE: roles_file,
            constants.OVERCLOUD_NETWORKS_FILE: networks_file,
        }
        # Also try to fetch any files under 'user-files/'
        # dir. local_name is always None for these
        keep_map.update(dict(map(
            lambda path: (path, None),
            _list_user_files(swift_client, name))))
        keep_file_contents = _load_content_or_file(
            swift_client, name, keep_map)
    else:
        passwords = _load_passwords(swift_client, name)

    # TODO(dmatthews): Removing the existing plan files should probably be
    #                  a Mistral action.
    print("Removing the current plan files")
    swiftutils.empty_container(swift_client, name)

    # Until we have a well defined plan update workflow in
    # tripleo-common we need to manually reset the environments and
    # parameter_defaults here. This is to ensure that no environments
    # are in the plan environment but not actually in swift.
    # See bug: https://bugs.launchpad.net/tripleo/+bug/1623431
    #
    # Currently this is being done incidentally because we overwrite
    # the existing plan-environment.yaml with the skeleton one in THT
    # when updating the templates. Once LP#1623431 is resolved we may
    # need to special-case plan-environment.yaml to avoid this.

    print("Uploading new plan files")
    if keep_env:
        _upload_templates(swift_client, name, tht_root)
        for filename in keep_file_contents:
            _upload_file_content(swift_client, name, filename,
                                 keep_file_contents[filename])
    else:
        _upload_templates(swift_client, name, tht_root, roles_file,
                          plan_env_file, networks_file)
        _update_passwords(swift_client, name, passwords)

    update_deployment_plan(
        clients, container=name,
        generate_passwords=generate_passwords,
        verbosity_level=verbosity_level,
        disable_image_params_prepare=disable_image_params_prepare)


def _load_content_or_file(swift_client, container, remote_and_local_map):
    # mapping (remote_name, content)
    file_contents = {}

    plan_files = _list_plan_files(swift_client, container)

    for remote_name in remote_and_local_map:
        LOG.debug("Attempting to load {0}".format(remote_name))
        local_name = remote_and_local_map[remote_name]
        # it's possible that the file doesn't exist in Swift and isn't
        # passed on filesystem, in which case we won't do anything
        content = None
        # local override takes priority
        if local_name:
            LOG.debug("Using provided file {0}".format(local_name))
            with open(os.path.abspath(local_name)) as local_content:
                content = local_content.read()
        elif remote_name in plan_files:
            LOG.debug("Preserving plan file {0}".format(remote_name))
            content = swift_client.get_object(container, remote_name)[1]

        if content:
            file_contents[remote_name] = content

    return file_contents


def _list_user_files(swift_client, container):
    return list(filter(lambda path: path.startswith('user-files/'),
                       _list_plan_files(swift_client, container)))


def _list_plan_files(swift_client, container):
    return list(map(lambda i: i['name'],
                    swift_client.get_container(
                        container, full_listing=True)[1]))


def _upload_file(swift_client, container, filename, local_filename):
    with open(local_filename, 'rb') as file_content:
        swiftutils.put_object_string(
            swift=swift_client,
            container=container,
            object_name=filename,
            contents=file_content
        )


# short function, just alias for interface parity with _upload_plan_file
def _upload_file_content(swift_client, container, filename, content):
    LOG.debug("Uploading {0} to plan".format(filename))
    swiftutils.put_object_string(
        swift=swift_client,
        container=container,
        object_name=filename,
        contents=content
    )


def _load_passwords(swift_client, name):
    plan_env = yaml.safe_load(swift_client.get_object(
        name, constants.PLAN_ENVIRONMENT)[1])

    if "passwords" in plan_env:
        return plan_env['passwords']
    else:
        LOG.warning("No passwords found in existing plan {}. "
                    "Updating plan with passwords.".format(name))


def _update_passwords(swift_client, name, passwords):
    # Update the plan environment with the generated passwords. This
    # will be solved more elegantly once passwords are saved in a
    # separate environment (https://review.opendev.org/#/c/467909/)
    if passwords:
        try:
            env = yaml.safe_load(swift_client.get_object(
                name, constants.PLAN_ENVIRONMENT)[1])
            env['passwords'] = passwords
            swiftutils.put_object_string(
                swift=swift_client,
                container=name,
                object_name=constants.PLAN_ENVIRONMENT,
                contents=yaml.safe_dump(env, default_flow_style=False)
            )
        except Exception as exp:
            # The plan likely has not been migrated to using Swift yet.
            LOG.debug("Could not find plan environment %s in %s - error: %s",
                      constants.PLAN_ENVIRONMENT, name, exp)


def export_deployment_plan(clients, plan_name):

    plan_path = os.path.join('/var/lib/tripleo/stacks', plan_name)
    if os.path.exists(plan_path):
        print("Plan information can be found here: {}".format(plan_path))
        return plan_path
    else:
        # NOTE(cloudnull): When swift is no longer in service remove this.
        export_container = "plan-exports"
        delete_after = 3600

        mistral_context = clients.tripleoclient.create_mistral_context()
        action = plan.ExportPlanAction(plan_name, delete_after=delete_after,
                                       exports_container=export_container)
        result = action.run(mistral_context)
        if result:
            raise exceptions.WorkflowServiceError(
                'Exception exporting plan: {}'.format(result.error))

        url = swiftutils.get_temp_url(
            clients.tripleoclient.object_store,
            container=export_container,
            object_name="{}.tar.gz".format(plan_name)
        )
        print(url)
        return url



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""OpenStackClient Plugin interface"""

import logging

from osc_lib import utils
from swiftclient import client as swift_client

LOG = logging.getLogger(__name__)

DEFAULT_TRIPLEOCLIENT_API_VERSION = '2'

# Required by the OSC plugin interface
API_NAME = 'tripleoclient'
API_VERSION_OPTION = 'os_tripleoclient_api_version'
API_VERSIONS = {
    '2': 'tripleoclient.plugin'
}


def make_client(instance):
    return ClientWrapper(instance)


# Required by the OSC plugin interface
def build_option_parser(parser):
    """Hook to add global options

    Called from openstackclient.shell.OpenStackShell.__init__()
    after the builtin parser has been initialized.  This is
    where a plugin can add global options such as an API version setting.

    :param argparse.ArgumentParser parser: The parser object that has been
        initialized by OpenStackShell.
    """
    parser.add_argument(
        '--os-tripleoclient-api-version',
        metavar='<tripleoclient-api-version>',
        default=utils.env(
            'OS_TRIPLEOCLIENT_API_VERSION',
            default=DEFAULT_TRIPLEOCLIENT_API_VERSION),
        help='TripleO Client API version, default=' +
             DEFAULT_TRIPLEOCLIENT_API_VERSION +
             ' (Env: OS_TRIPLEOCLIENT_API_VERSION)')
    return parser


class MistralContext(object):
    """MistralContext, a shim for calling Mistral actions directly

    The MistralContext and MistralSecurityContext combined mimic the context
    which Mistral passes to actions during a Workflow execution. It does
    not include all the data or cover all of the functionality but it does
    include everything we use in tripleo-common.

    The MistralContext should be created by the create_mistral_context method
    on the ClientWrapper class below.

    This should be refactored and removed once Mistral server has been removed.
    """
    def __init__(self, security_ctx=None):
        self.security = security_ctx

    def __getattribute__(self, name):
        deprecated = [
            "auth_cacert", "auth_token", "auth_uri", "expires_at", "insecure",
            "is_target", "is_trust_scoped", "project_id", "project_name",
            "redelivered", "region_name", "service_catalog", "trust_id",
            "user_name"
        ]
        if name in deprecated:
            return getattr(self.security, name)
        return super(MistralContext, self).__getattribute__(name)


class MistralSecurityContext(object):
    def __init__(self, auth_uri=None, auth_cacert=None, insecure=None,
                 service_catalog=None, region_name=None, is_trust_scoped=None,
                 redelivered=None, expires_at=None, trust_id=None,
                 is_target=None, project_id=None, project_name=None,
                 user_name=None, user_id=None, auth_token=None):
        self.auth_uri = auth_uri
        self.auth_cacert = auth_cacert
        self.insecure = insecure
        self.service_catalog = service_catalog
        self.region_name = region_name
        self.is_trust_scoped = is_trust_scoped
        self.redelivered = redelivered
        self.expires_at = expires_at
        self.trust_id = trust_id
        self.is_target = is_target
        self.project_id = project_id
        self.project_name = project_name
        self.user_name = user_name
        self.user_id = user_id
        self.auth_token = auth_token


class ClientWrapper(object):

    def __init__(self, instance):
        self._instance = instance
        self._object_store = None
        self._local_orchestration = None

    def create_mistral_context(self):
        """Create a Mistral context

        Create a class that mimics the Mistral context. This allows us to call
        Mistral action classes directly.

        See the docstring on MistralContext for more context.
        """
        session = self._instance.session
        security_ctx = MistralSecurityContext(
            auth_token=self._instance.auth.get_token(session),
            auth_uri=self._instance.auth.auth_url,
            project_id=self._instance.auth.get_project_id(session),
            project_name=self._instance.auth._project_name,
            service_catalog=session.auth.auth_ref._data['token'],
            trust_id=self._instance.auth_ref.trust_id,
            user_name=self._instance.auth._username,
            auth_cacert=self._instance.cacert,
            user_id=self._instance.auth._user_id
        )
        return MistralContext(security_ctx=security_ctx)

    def local_orchestration(self, api_port):
        """Returns an local_orchestration service client"""

        if self._local_orchestration is not None:
            return self._local_orchestration

        API_VERSIONS = {
            '1': 'heatclient.v1.client.Client',
        }

        heat_client = utils.get_client_class(
            API_NAME,
            '1',
            API_VERSIONS)
        LOG.debug('Instantiating local_orchestration client: %s', heat_client)

        client = heat_client(
            endpoint='http://127.0.0.1:%s/v1/admin' % api_port,
            username='admin',
            password='fake',
            region_name='regionOne',
            token='fake',
        )

        self._local_orchestration = client
        return self._local_orchestration

    @property
    def object_store(self):
        """Returns an object_store service client

        The Swift/Object client returned by python-openstack client isn't an
        instance of python-swiftclient, and had far less functionality.
        """

        if self._object_store is not None:
            return self._object_store

        endpoint = self._instance.get_endpoint_for_service_type(
            "object-store",
            region_name=self._instance._region_name,
        )

        token = self._instance.auth.get_token(self._instance.session)

        kwargs = {
            'preauthurl': endpoint,
            'preauthtoken': token
        }

        self._object_store = swift_client.Connection(**kwargs)
        return self._object_store



#   Copyright 2016 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import logging
import os

from osc_lib.i18n import _
import yaml

from tripleoclient import command
from tripleoclient import utils
from tripleoclient.workflows import baremetal


class CreateRAID(command.Command):
    """Create RAID on given nodes"""

    log = logging.getLogger(__name__ + ".CreateRAID")

    def get_parser(self, prog_name):
        parser = super(CreateRAID, self).get_parser(prog_name)
        parser.add_argument('--node', action='append', required=True,
                            help=_('Nodes to create RAID on (expected to be '
                                   'in manageable state). Can be specified '
                                   'multiple times.'))
        parser.add_argument('configuration',
                            help=_('RAID configuration (YAML/JSON string or '
                                   'file name).'))
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action({args})".format(args=parsed_args))

        if os.path.exists(parsed_args.configuration):
            with open(parsed_args.configuration, 'r') as fp:
                configuration = yaml.safe_load(fp.read())
        else:
            try:
                configuration = yaml.safe_load(parsed_args.configuration)
            except yaml.YAMLError as exc:
                raise RuntimeError(
                    _('Configuration is not an existing file and cannot be '
                      'parsed as YAML: %s') % exc)

        # Basic sanity check, we defer the full check to Ironic
        try:
            disks = configuration['logical_disks']
        except KeyError:
            raise ValueError(
                _('Configuration must contain key "logical_disks"'))
        except TypeError:
            raise TypeError(
                _('Configuration must be an object, got %r instead')
                % configuration)

        if (not isinstance(disks, list) or
                not all(isinstance(item, dict) for item in disks)):
            raise TypeError(
                _('Logical disks list is expected to be a list of objects, '
                  'got %r instead') % disks)

        baremetal.create_raid_configuration(
            clients=self.app.client_manager,
            node_uuids=parsed_args.node,
            configuration=configuration,
            verbosity=utils.playbook_verbosity(self=self)
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from datetime import datetime
import logging
import os.path
import yaml

from osc_lib.i18n import _
from osc_lib import utils

from tripleoclient import command
from tripleoclient import export


class ExportOvercloud(command.Command):
    """Export stack information used as import of another stack"""

    log = logging.getLogger(__name__ + ".ExportOvercloud")
    now = datetime.now().strftime('%Y%m%d%H%M%S')

    def get_parser(self, prog_name):
        parser = super(ExportOvercloud, self).get_parser(prog_name)
        parser.add_argument('--stack',
                            dest='stack',
                            metavar='<stack>',
                            help=_('Name of the environment main Heat stack '
                                   'to export information from. '
                                   '(default=Env: OVERCLOUD_STACK_NAME)'),
                            default=utils.env('OVERCLOUD_STACK_NAME',
                                              default='overcloud'))
        parser.add_argument('--output-file', '-o', metavar='<output file>',
                            help=_('Name of the output file for the stack '
                                   'data export. It will default to '
                                   '"<name>.yaml"'))
        parser.add_argument('--force-overwrite', '-f', action='store_true',
                            default=False,
                            help=_('Overwrite output file if it exists.'))
        parser.add_argument('--config-download-dir',
                            action='store',
                            help=_('Directory to search for config-download '
                                   'export data. Defaults to '
                                   '$HOME/config-download'))
        parser.add_argument('--no-password-excludes',
                            action='store_true',
                            dest='no_password_excludes',
                            help=_('Don''t exclude certain passwords from '
                                   'the password export. Defaults to False '
                                   'in that some passwords will be excluded '
                                   'that are not typically necessary.'))

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        stack = parsed_args.stack
        output_file = parsed_args.output_file or \
            '%s-export.yaml' % stack

        self.log.info('Running at %s with parameters %s',
                      self.now,
                      parsed_args)

        if os.path.exists(output_file) and not parsed_args.force_overwrite:
            raise Exception(
                "File '%s' already exists, not exporting." % output_file)

        if not parsed_args.config_download_dir:
            config_download_dir = os.path.join(os.environ.get('HOME'),
                                               'config-download')
        else:
            config_download_dir = parsed_args.config_download_dir

        # prepare clients to access the environment
        clients = self.app.client_manager
        swift_client = clients.tripleoclient.object_store

        data = export.export_passwords(swift_client, stack,
                                       not parsed_args.no_password_excludes)
        data.update(export.export_stack(
            clients.orchestration, stack, False, config_download_dir))
        data = dict(parameter_defaults=data)

        # write the exported data
        with open(output_file, 'w') as f:
            yaml.safe_dump(data, f, default_flow_style=False)

        print("Stack information exported to %s." % output_file)
